/* jshint undef: true */
/* global plugin, require, com, java, javax, org */

var _ = require("lodash");
var on = require("on")(plugin);
var Objects = require("convenience").Objects;
var getStackTraceAsString = require("convenience").Errors.getStackTraceAsString;
var facade = require("facade");
var SystemInformationUtilities = com.cattura.agents.configuration_module.model.utilities.SystemInformationUtilities;
var StringUtilities = com.cattura.cattura_commons.utilities.StringUtilities;
var EventType = com.cattura.agents.configuration_module.model.vos.events.EventType;
var EmailBuilder = org.simplejavamail.email.EmailBuilder;
var Mailer = org.simplejavamail.mailer.Mailer;
var FileDataSource = javax.activation.FileDataSource;
var STORE = plugin.info().dataStore;
var settings = plugin.info().settings;
var LOGGER = plugin.info().logger;
var StringWriter = java.io.StringWriter;
var createAttachmentForIncident = require("./createAttachmentForIncident");

var Level = java.util.logging.Level;

var ThreadFactoryBuilder = com.google.common.util.concurrent.ThreadFactoryBuilder;
var successfulAsList = com.google.common.util.concurrent.Futures.successfulAsList;
var listeningDecorator = com.google.common.util.concurrent.MoreExecutors.listeningDecorator;
var newCachedThreadPool = java.util.concurrent.Executors.newCachedThreadPool;

/*
 * Executor used for sending emails
 */
var mailExecutor = listeningDecorator(newCachedThreadPool(new ThreadFactoryBuilder()
    .setNameFormat("Email Notifier Plugin executor thread %d")
    .setDaemon(true)
.build()));

/*
 * This plugin will require an SMTP server to talk to.
 * By default it will use localhost; thus, assuming that
 * an SMTP server is running on the recorder. The standard
 * distribution comes with OpenSMTPd.
 */
var mailer = new Mailer(
    settings.getString("host").or("localhost"),   /* SMTP host */
    settings.getString("port").or(25).intValue(), /* SMTP port */
    settings.getString("username").orNull(),      /* Username, not needed */ 
    settings.getString("password").orNull()       /* Password, not needed */
);

function Notifier (data)
{
    _(this).extend(data).value();
    
    this.canHandleEvent = function canHandleEvent (event)
    {
        if (event.type.category == EventType.Category.INTERNAL)
        {
            return false;
        }
        
        return _(this.events).find(function (type) { 
            return type == event.type.name() || type == event.getClass().simpleName;
        });
    };
    
    this.handleEvent = function handleEvent (event, attachments)
    {
        var t = this;
        return mailExecutor['submit(java.lang.Runnable)'](function () {
            var builder = new EmailBuilder().from("Cattura", "cattura@" + SystemInformationUtilities.getMachineName())
                                            .to(t.email, t.email)
                                            .subject("[" + getUnitName() + "] " + event.description)
                                            .text(event.extendedDescription || event.description);
            
            if (event.type.name() == "EXCEPTION")
            {
                if (event.incident)
                {
                    attachExceptionText(event, builder);
                }
                else
                {
                    return; // don't want to handle this if it's a legacy ExceptionEvent
                }
            }
            
            addAttachments(builder, attachments);
            mailer.sendMail(builder.build());
        });
    };
    
    function addAttachments (builder, attachments)
    {
        if (!attachments)
        {
            return;
        }
        
        for (var filename in attachments)
        {
            try
            {
                builder.addAttachment(filename, new FileDataSource(attachments[filename]));
            }
            catch (e)
            {
                LOGGER.warning("Failed to attach " + filename + " to event report");
            }            
        }
    }
    
    function attachExceptionText (event, builder)
    {
        if (event.incident.exception)
        {
            builder.text(getStackTraceAsString(event.incident.exception));
        }
    }    
}

function getNotifiers ()
{
    return _(Objects.toJS(STORE.select("notifier:*"))).map(function (notifier) {
        return new Notifier(notifier);
    }).value();
}

function addNotifier (id, email, events)
{
    STORE.set("notifier:" + id, {id: id, email: email, events: events});
}

function removeNotifier (id)
{
    STORE.remove("notifier:" + id);
}

function getNotifiersForEvent (event)
{
    return _(getNotifiers()).filter(function (notifier) {
        return notifier.canHandleEvent(event);
    }).value();
}

function generateIdForNotifier ()
{
    var id = StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");
    if (_(getNotifiers()).find(function (notifier) { return notifier.id === id; }))
    {
        return generateIdForNotifier();
    }
    
    return id;
}

function getUnitName ()
{
    var settings = facade.retrieveProxy("SettingsProxy");
    var unitName = settings.getUserSettings().getString("unit-name").or("");
    if (!unitName.length())
    {
        unitName = SystemInformationUtilities.getMachineName();
    }
    return unitName;
}

function getExceptionFromEvent (event)
{
    if ('incident' in event)
    {
        return event.incident.exception;
    }
    else if ('exception' in event)
    {
        return event.exception;
    }
}

function generateAttachmentsForEvent (event)
{
    var attachments = {};
    if (!(event.type.name() == "EXCEPTION" && event.incident))
    {
        return attachments;
    }
    
    try
    {
        var attachment = createAttachmentForIncident();
        if (attachment)
        {
            attachments["catturalogs.zip"] = attachment;
        }
    }
    catch (e)
    {
        LOGGER.warning("Failed to generate attachment for error report: " + getStackTraceAsString(e));
    }
    
    return attachments;
}

function cleanupAttachments (attachments)
{
    if (!attachments)
    {
        return;
    }
    
    _(attachments).each(function (attachment) {
        attachment.delete();
    });
}

plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "email", description: "Email address", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"}
    ],
    handler: function (email, events) {
        addNotifier(generateIdForNotifier(), email, events);
    }
});

plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "id", description: "Unique identifier for notifier", type: "java.lang.String"},
        {name: "email", description: "Email address", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"}
    ],
    handler: addNotifier
});

plugin.service({
    name: "removeNotifier",
    description: "Remove a notifier",
    parameters: [
        {name: "id", description: "Unique identifier for notifier", type: "java.lang.String"}
    ],
    handler: removeNotifier
});

plugin.service({
    name: "getNotifiers",
    description: "Return all registered notifiers",
    parameters: [],
    handler: getNotifiers
});

function handleExceptionFromHandler (event, exception)
{
    var unwrappedException = exception.javaException || exception.rhinoException || exception;
    if (getExceptionFromEvent(event))
    {
        LOGGER.log(Level.WARNING, "Email Notification Plugin encountered (and swallowed) exception while trying to handle exception event", unwrappedException);
        return;
    }
    
    facade.sendNotification("HANDLE_EXCEPTION", new java.lang.RuntimeException("Email Notification Plugin encountered exception while trying to handle event: " + event, unwrappedException));
}

plugin.onEvent(function(event) {
    try
    {
        var notifiers = getNotifiersForEvent(event);
        if (!notifiers.length)
        {
            return;
        }
        
        /*
         * Create attachments and transmit events using executor.
         * Emails are sent in parallel, then when all emails are sent
         * the attachments are cleaned up.
         */
        mailExecutor.execute(function () {
            try
            {
                var attachments = generateAttachmentsForEvent(event);
                successfulAsList(_(notifiers).map(function (notifier) {
                    return notifier.handleEvent(event, attachments);
                }).value()).addListener(function () {
                    cleanupAttachments(attachments);
                }, mailExecutor);
            }
            catch (exception)
            {
                handleExceptionFromHandler(event, exception);
            }
        });
    }
    catch (exception)
    {
        handleExceptionFromHandler(event, exception);
    }
});